# -*- Encoding: UTF-8 -*-
###
# Copyright (c) 2008-2009, Elián Hanisch
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###

# imports
import supybot.log as log
import supybot.conf as conf
import supybot.utils as utils
from supybot.commands import *
import supybot.plugins as plugins
import supybot.ircutils as ircutils
import supybot.callbacks as callbacks
from supybot.ircutils import IrcString

import re, time

def now():
	return int(time.time())

def strTime(t):
	"""return a readable date from an int time."""
	return time.strftime('%d/%m/%y', time.localtime(int(t)))

# XXX retrieve this from config
tblFactName = IrcString('factos')
tblHistoryName = IrcString('historial')
tblUserName = IrcString('usuarios')
# table definitions
tblFactFields = (
		('key', 'TEXT PRIMARY KEY'),
		('created_by', 'INTEGER'),
		('created_at', 'TIMESTAMP'),
		('edited_by', 'INTEGER'),
		('edited_at', 'TIMESTAMP'),
		('flags', 'INTEGER'),
		('flags_set_by', 'INTEGER'),
		('flags_set_at', 'TIMESTAMP'),
		('last_request_by', 'TEXT'),
		('last_request_at', 'TIMESTAMP'),
		('request_count', 'INTEGER'),
		('fact', 'TEXT'),
		)
tblHistoryFields = (
		('key', 'TEXT'),
		('history_id', 'INTEGER'),
		('edited_by', 'INTEGER'),
		('edited_at', 'TIMESTAMP'),
		('fact', 'TEXT'),
		)
tblUserFields = (
		('user_id', 'INTEGER PRIMARY KEY'),
		('host', 'TEXT'),
		('first_seen', 'TIMESTAMP'),
		('last_seen', 'TIMESTAMP'),
		('facts', 'INT'),
		('edits', 'INT'),
		)

# I need something like a dictionary with ordenable items
class table(tuple):

	def __new__(cls, values, name=None):
		return tuple.__new__(cls, values)

	def __init__(self, values, name):
		self.baseName = name

	def __getitem__(self, keyword):
		try:
			return tuple.__getitem__(self, self.index(keyword))[1]
		except TypeError:
			return None

	def index(self, keyword):
		list = [ k for (k, v) in self ]
		try:
			return list.index(keyword)
		except:
			return None

# setup the tables types we will use
tableFact = table(tblFactFields, tblFactName)
tableHistory = table(tblHistoryFields, tblHistoryName)
tableUser = table(tblUserFields, tblUserName)

# special tuple we will use for get and insert data in the database
class row(tuple):

	def __new__(cls, tlbDef, values):
		assert len(tlbDef) == len(values), \
				'Values count doesn\'t match, expected %d' % len(tlbDef)
		return tuple.__new__(cls, values)

	def __init__(self, tlbDef, values):
		self.header = [ k for (k, v) in tlbDef ]
		self.table = tlbDef

	def __getitem__(self, keyword):
		try:
			return tuple.__getitem__(self, self.index(keyword))
		except TypeError:
			return None

	def index(self, keyword):
		try:
			return self.header.index(keyword)
		except:
			return None

	def repr(self):
		"""Returns a row suitable for using in SQL syntax"""
		list = []
		datatypeIter = [ v for (k, v) in self.table ].__iter__()
		for v in self:
			datatype = datatypeIter.next()
			if v == None:
				list.append('NULL')
			elif datatype.startswith('TEXT'):
				v = str(v).replace('\'', '\'\'')
				list.append('\'%s\'' % v)
			else:
				list.append(str(v))
		return row(self.table, list)

# row for fact's table
class factRow(row):

	def __new__(cls, values):
		return row.__new__(cls, tableFact, values)

	def __init__(cls, values):
		row.__init__(cls, tableFact, values)

# row for user's table
class userRow(row):

	def __new__(cls, values):
		return row.__new__(cls, tableUser, values)

	def __init__(cls, values):
		row.__init__(cls, tableUser, values)

# row for history's table
class historyRow(row):

	def __new__(cls, values):
		return row.__new__(cls, tableHistory, values)

	def __init__(cls, values):
		row.__init__(cls, tableHistory, values)


class DataObject:
	fieldsDB = ()

	def setAllFields(self, values):
		iter = values.__iter__()
		for (k, n) in self.fieldsDB:
			setattr(self, k, iter.next())

	def getAllFields(self):
		all = []
		for (k, n) in self.fieldsDB:
			all.append(getattr(self, k))
		return row(self.fieldsDB, all)

	def update(self, dict):
		for (k, v) in dict.iteritems():
			setattr(self, k, v)

	def raw(self):
		s = ' '.join([ '%s:%s' % (k, getattr(self, k)) for (k, n) in
			self.fieldsDB ])
		return s

# class for handling facts.
# XXX maybe merge it with factRow class?
class Fact(DataObject):
	keyMaxLen = 25 # XXX make configurable
	factMaxLen = 400 # XXX make configurable
	fieldsDB = tableFact
	# status flags
	f = {'disable':1, 'lock':2, 'mask':3}

	def __init__(self, key='', inChannel=None, **kwargs):
		self.user = User()
		self.db = None
		self.channel = None
		self.inChannel = inChannel
		self.table = None
		self.options = ()
		self.toNick = None
		self.private = False
		self.key = ircutils.toLower(key)
		self.created_by = None
		self.created_at = now()
		self.edited_by = None
		self.edited_at = None
		self.flags = 0
		self.flags_set_by = None
		self.flags_set_at = None
		self.last_request_by = None
		self.last_request_at = None
		self.request_count = 0
		self.fact = ''
		for (k, v) in kwargs.iteritems():
			setattr(self, k, v)

	def update(self, dict):
		if dict.has_key('key'):
			dict['key'] = ircutils.toLower(dict['key'])
		DataObject.update(self, dict)

	def edit(self, value):
		self.old_fact = self.copy()
		self.fact = value
		self.user.pull()
		self.edited_by = self.user.user_id
		self.edited_at = now()
		self.user.countEdits()
		self.user.push()

	def setLastRequest(self):
		self.count()
		self.last_request_by = self.user.host
		self.last_request_at = now()

	def copy(self):
		copy = Fact(self.key)
		for attr in [ a for a in dir(self) if not callable(getattr(self, a))
				and a[0] != '_' ]:
			setattr(copy, attr, getattr(self, attr))
		return copy

	def isDisabled(self):
		if self.flags & self.f['disable']:
			return True
		return False

	def isLocked(self):
		if self.flags & self.f['lock']:
			return True
		return False

	def isAlias(self):
		if self.fact.startswith('<alias>'):
			return True
		else:
			return False

	def setToAlias(self):
		if self.isAlias():
			self.key = self.fact[len('<alias>'):]
			self.channel = None

	def count(self):
		self.request_count += 1

	def disable(self):
		self.setFlag('+disable')

	def undisable(self):
		self.setFlag('-disable')

	def lock(self):
		self.setFlag('+lock')

	def unlock(self):
		self.setFlag('-lock')

	def setFlag(self, flag):
		self.flags_set_by = self.user.user_id
		self.flags_set_at = now()
		(action, flag) = (flag[0], flag[1:])
		if action == '+':
			self.flags = self.flags | self.f[flag]
		elif action == '-':
			self.flags = self.flags & ( self.f[flag] ^ self.f['mask'] )

	def add(self):
		self.user.pull()
		self.created_by = self.user.user_id
		self.user.countFacts()
		self.user.push()
		values = self.getAllFields()
		self.db.addFact(values, self.inChannel, self.channel)

	def pull(self, pullDisabled = False):
		"""Pulls data from the database, if the database doesn't have a matching
		fact, self.fact is set to '' hence fact.isValid() will return False."""
		log.debug('Fact.pull()')
		tableList = self.db.getTables(self.inChannel, self.channel)
		for (table, channel) in tableList:
			values = self.db.getFact(self.key, table, self.inChannel,
					field='key')
			if values:
				self.setAllFields(values)
				self.table = table
				self.channel = channel
				if pullDisabled:
					return
				else:
					if not self.isDisabled():
						return
		self.invalidate()

	def push(self):
		"""Push data to the database, if the database doesn't have a matching
		fact... this will probably fail somewhere."""
		# XXX add assertion somewhere for this ^
		log.debug('Fact.push()')
		values = self.getAllFields()
		self.db.upFact(values, self.table, self.inChannel)
		if hasattr(self, 'old_fact'):
			if self.fact != self.old_fact.fact:
				self.db.addHistory(self.old_fact.getAllFields(),
						self.inChannel, self.channel)

	def __str__(self):
		return self.fact[:self.factMaxLen]

	def __repr__(self):
		fact = self.fact.replace('\'', '\'\'')
		return '\'%s\'' % fact[:self.factMaxLen]

	def __len__(self):
		return len(self.fact)

	def isValid(self):
		if self.user.host is None:
			return False
		if len(self) > self.factMaxLen:
			return False
		if len(self.key) > self.keyMaxLen:
			return False
		return len(self)

	def invalidate(self):
		self.fact = ''

# class for handling users.
# XXX maybe merge it with userRow class?
class User(DataObject):

	def __init__(self, host=None, inChannel=None, user_id=None, **kwargs):
		self.fieldsDB = tableUser
		self.inChannel = inChannel
		self.db = None
		self.host = host
		self.user_id = user_id
		self.first_seen = now()
		self.last_seen = self.first_seen
		self.facts = 0
		self.edits = 0
		for (opt, arg) in kwargs.iteritems():
				setattr(self, opt, arg)

	def push(self):
		"""Push data to the database, if the database doesn't have a matching
		user... this will probably fail somewhere."""
		# XXX add assertion somewhere for this ^
		log.debug('User.push()')
		values = self.getAllFields()
		self.db.upUser(values, self.inChannel)

	def pull(self):
		"""Pulls data from the database, if the database doesn't have a matching
		user and self.host is supplied, it will be created."""
		log.debug('User.pull()')
		if not self.host:
			assert self.user_id != None, 'no user id supplied'
			values = self.db.findUser(self.user_id, self.inChannel)
			#log.debug('User.pull() findUser values=%s' % repr(values))
			self.setAllFields(values)
		else:
			values = self.db.getUser(self.host, self.inChannel)
			#log.debug('User.pull() getUser values=%s' % repr(values))
			self.setAllFields(values)
			self.upSeen()

	def upSeen(self):
		self.last_seen = now()

	def countFacts(self):
		self.facts += 1

	def countEdits(self):
		self.edits += 1

	def getNick(self):
		# XXX this should return ircstring
		if self.host:
			return self.host.split('!')[0]

	def __str__(self):
		# XXX return self.getNick would be more usefull?
		return self.host

	def __len__(self):
		return len(self.host)

# class for manipulate sqlite databases
class SqliteDB:

	def __init__(self, filename):
		self.filename = filename
		self.dbs = ircutils.IrcDict()

	def close(self):
		for db in self.dbs.itervalues():
			db.close()
		self.dbs.clear()

	def commit(self, db, cmd):
		log.debug('SqliteDB.commit() cmd=%s' % repr(cmd))
		cursor = db.cursor()
		try:
			cursor.execute(cmd)
		except Exception, e:
			db.rollback()
			raise callbacks.Error, e.message
		db.commit()
		cursor.close()

	def __getDb(self, channel):
		"""Returns database name for channel, and creates if it doesn't
		exits"""

		def connect(filename):
			return sqlite.connect(filename, encoding='utf-8')

		try:
			import sqlite
		except ImportError:
			raise callbacks.Error, \
					'You need to have PySQLite installed to use this ' \
					'plugin.  Download it at <http://pysqlite.org/>'
		# XXX make channel specific plugin configurable
		# independent of supybot.databases.plugins.channelSpecific
		channel = '#'
		if channel in self.dbs:
			return self.dbs[channel]
		filename = plugins.makeChannelFilename(self.filename, channel)
		db = connect(filename)
		db.tables = []
		db.users = {}
		self.dbs[channel] = db
		return db

	def __getTable(self, db, tblType, channel=None, create=False):
		"""returns table name according to table type and channel, creates it if
		create is True, otherwise returns None"""

		def createTable(name, type):
			fields = ', '.join([ '%s %s' % (k, v) for (k, v) in type ])
			cmd = """CREATE TABLE '%s' ( %s )""" % (name, fields)
			self.commit(db, cmd)

		def checkTable(name, type):
			cursor = db.cursor()
			try:
				cursor.execute("""SELECT sql FROM sqlite_master WHERE
				name=%s""", name)
				table = cursor.fetchall()[0][0]
			except Exception, e:
				return False
			cursor.close()
			found = table[table.rfind('('):].strip('() ')
			expect = ', '.join([ '%s %s' % (k, v) for (k, v) in tblType ])
			if expect != found:
				raise callbacks.Error, 'La tabla \'%s\' no es la esperada' % \
						name
			return True

		# if we actually want the global table
		if channel == '#':
			channel = None
		if (tblType.baseName != tableUser.baseName) and channel:
			tblName = '%s_%s' % (tblType.baseName, channel)
		else:
			tblName = tblType.baseName
		tblName = IrcString(ircutils.toLower(tblName))
		if tblName in db.tables:
			return tblName
		if checkTable(tblName, tblType):
			db.tables.append(tblName)
			return tblName
		if not create:
			return None
		log.debug('SqliteDB.__getTable() create channel=%s' % repr(channel))
		createTable(tblName, tblType)
		db.tables.append(tblName)
		return tblName

	def getTables(self, inChannel, channel=None):
		"""returns a list of available tables, if channel defined returns the
		table of that channel, else checks the local table to inChannel
		and the global table."""
		log.debug('SqliteDB.getTables() inChannel=%s channel=%s' %
				(repr(inChannel), repr(channel)))
		db = self.__getDb(inChannel)
		tableList = []
		if channel:
			# if we actually want the global table
			# if channel == '#':
			# 	channel = None
			tbl = self.__getTable(db, tableFact, channel)
			if tbl:
				tableList = [(tbl, channel), ]
		else:
			localTable = self.__getTable(db, tableFact, inChannel)
			globalTable = self.__getTable(db, tableFact)
			tableList = ((localTable, inChannel), (globalTable, None))
			tableList = [ t for t in tableList if t[0] ]
		return tableList

	def addFact(self, values, inChannel, channel):
		"""adds fact in database"""
		db = self.__getDb(inChannel)
		key = values['key']
		tbl = self.__getTable(db, tableFact, channel, create=True)
		fields = ', '.join(values.repr())
		cmd = """INSERT INTO '%s' VALUES ( %s )""" % (tbl, fields)
		self.commit(db, cmd)

	def getFact(self, key, table, inChannel, field):
		"""returns tuple from database"""
		db = self.__getDb(inChannel)
		values = ()
		cursor = db.cursor()
		cmd = """SELECT * FROM '%s' WHERE %s LIKE '%s'""" % (table, field, key)
		cursor.execute(cmd)
		if cursor.rowcount > 0:
			values = factRow(cursor.fetchall()[0])
		cursor.close()
		return values

	def delFact(self, key, table, inChannel):
		"""deletes fact from database"""
		db = self.__getDb(inChannel)
		cursor = db.cursor()
		cmd = """DELETE FROM '%s' WHERE key LIKE %s""" % (table, key)
		self.commit(db, cmd)

	def upFact(self, values, table, inChannel, fields=None):
		"""Updates fact in database and adds an history entry if fact.fact
		changed"""
		db = self.__getDb(inChannel)
		fields = [ '%s=%s' % (k, values.repr()[k]) for k in values.header ]
		fields = ', '.join(fields)
		cmd = """UPDATE '%s' SET %s WHERE key LIKE '%s'""" % (table,
				fields, values['key'])
		self.commit(db, cmd)

	def addHistory(self, values, inChannel, channel):
		db = self.__getDb(inChannel)
		key = values['key']
		tbl = self.__getTable(db, tableHistory, channel, create=True)
		cursor = db.cursor()
		cmd = """SELECT * FROM '%s' WHERE key LIKE '%s'""" % (tbl, key)
		cursor.execute(cmd)
		history_id = cursor.rowcount + 1
		cursor.close()
		values = historyRow((key, history_id, values['edited_by'],
			values['edited_at'], values['fact']))
		fields = ', '.join(values.repr())
		cmd = """INSERT INTO '%s' VALUES ( %s )""" % (tbl, fields)
		self.commit(db, cmd)

	def getUser(self, host, inChannel):
		"""Returns user from database or creates a new user entry and
		returns it if isn't in database"""
		db = self.__getDb(inChannel)
		tbl = self.__getTable(db, tableUser, create=True)
		if host in db.users.keys():
			values = db.users[host]
			return values
		cursor = db.cursor()
		cursor.execute("""SELECT * FROM %s WHERE host LIKE %s""",
				tbl, host)
		if cursor.rowcount > 0:
			values = userRow(cursor.fetchall()[0])
			db.users[host] = values
			cursor.close()
			return values
		cursor.execute("""SELECT * FROM %s""", tbl)
		user = User(host, user_id=cursor.rowcount)
		cursor.close()
		values = user.getAllFields()
		fields = ', '.join(values.repr())
		cmd = """INSERT INTO '%s' VALUES ( %s )""" % (tbl, fields)
		self.commit(db, cmd)
		db.users[host] = values
		return values

	def findUser(self, user_id, inChannel):
		db = self.__getDb(inChannel)
		tbl = self.__getTable(db, tableUser)
		cursor = db.cursor()
		cursor.execute("""SELECT * FROM %s WHERE user_id LIKE %s""",
				tbl, user_id)
		if cursor.rowcount > 0:
			values = userRow(cursor.fetchall()[0])
			return values
		else:
			raise callbacks.Error, 'User_id %s no encontrado' % user_id
		cursor.close()

	def upUser(self, values, inChannel):
		db = self.__getDb(inChannel)
		tbl = self.__getTable(db, tableUser)
		fields = [ '%s=%s' % (k, values.repr()[k]) for k in values.header ]
		fields = ', '.join(fields)
		cmd = """UPDATE '%s' SET %s WHERE user_id LIKE '%s'""" % (tbl, fields,
				values['user_id'])
		self.commit(db, cmd)
		db.users[values['host']] = values

	def search(self, keywords, inChannel):
		db = self.__getDb(inChannel)
		tables = self.getTables(inChannel)
		log.debug('SqliteDB.search() tables=%s' % repr(tables))
		conditions = ' AND '.join(['(key LIKE \'%%%s%%\' OR fact LIKE \'%%%s%%\')' % (k, k) for k in keywords])
		results = []
		cursor = db.cursor()
		for tbl in tables:
			cmd = 'SELECT key FROM \'%s\' WHERE %s' % (tbl[0], conditions)
			cursor.execute(cmd)
			if cursor.rowcount > 0:
				L = [ k[0] for k in cursor.fetchall() ]
				results.extend(L)
		cursor.close()
		return results


defaultTestMethods = (
		'getFact',
		'addFact',
		'editFact',
		'addAlias',
		)
defaultMethodArgs = {
		'getFact':('info', 'raw', 'pipe'),
		}
# class for parsing some of bot's commands
class ParseMsg:

	def __init__(self):
		validChars = r'\wñÑ\.\-\#'
		self.regexp = {
			'pipe':re.compile (
				r"""
				^
				(.+?)
				\s*(\||>)           # delimitador
				\s*([^\s]+)         # nombre del nick
				\s*$                # espacios al final""",
				re.VERBOSE),
			'getFact':re.compile (
				r"""
				^                   # principio
				([%s]+)             # una palabra
				\s*$                # espacios al final""" % validChars,
				re.VERBOSE),
			'addFact':re.compile (
				r"""
				^                   # principio
				([%s]+)             # una palabra si o si
				\s*(\bes\b|:)\s*    # delimitador, es o :
				(.+?)               # cualquier cosa (facto)
				\s*$                # espacios al final""" % validChars,
				re.VERBOSE),
			'editFact':re.compile (
				r"""
				^no,?\s*            # empezar con no o no,
				(.+)                # cualquier cosa
				\s*$""",
				re.VERBOSE),
			'addAlias':re.compile (
				r"""
				^
				([%s]+)
				\s*\balias\b
				\s*([%s]+)
				\s*$""" % (validChars, validChars),
				re.VERBOSE),
		}

	def parse(self, msg, testMethods=defaultTestMethods, methodArgs=defaultMethodArgs):
		"""This function will test a string for the methods in testMethods until it matchs a command.
		It will use the regexp in self.regexp for matching if it's defined, if is not it will directly
		call the method. These methods will update the dictionary self.parser which will returned by this
		function. If a method returns True it will stop searching. Methods that use a regexp start with a
		underscore.
		Some commands might accept specials options like a --flag, these are defined by methodArgs and
		handled by the self.options() function."""
		self.msg = msg
		self.parser = {}
		log.debug('ParseMsg.parse() testMethods=%s' % repr(testMethods))
		for m in testMethods:
			method = hasattr(self, '_' + m) and getattr(self, '_' + m) or None
			if method:
				match = None
				if methodArgs.has_key(m):
					msgBak = self.msg
					parserBak = self.parser.copy()
					log.debug('parseMsg.parse() parsing options() for %s' %	repr(m))
					self.options(methodArgs[m])
				try:
					match = self.regexp[m].search(self.msg)
				except:
					# always match if there's no test regexp
					match = True
				if match:
					log.debug('parseMsg.parse() method=%s msg=%s' %	(repr(m),repr(self.msg)))
					if method(match):
						self.parser['method'] = m
						log.debug('parseMsg.parse() return=%s' % repr(self.parser))
						return self.parser
				else:
					if methodArgs.has_key(m):
						# reverting possible changes made by options()
						self.msg = msgBak
						self.parser = parserBak

	def stripChannel(self, key):
		"""A fact is defined by a keyword and a channel's name: keyword#channel, this function splits them
		and returns a tuple.
		(key, channel) = stripChannel(fact)"""
		if '#' in key:
			(key, channel) = key.split('#')
			channel = '#' + channel
			#log.debug('paserMsg.stripChannel() key=%s channel=%s' % (repr(key), repr(channel)))
			return (key, channel)
		else:
			return (key, None)

	def options(self, args):
		"""Handles special options for each method."""
		msg = self.msg
		L = []
		for opt in args:
			if opt == 'pipe':
				parser = self.parser.copy()
				self.parse(msg, testMethods=('pipe', ))
				self.parser.update(parser)
				msg = self.msg
				continue
			s = '--' + opt
			if s in msg:
				L.append(opt)
				msg = msg.replace(s, '')
		self.msg = msg
		self.parser.update({'options':L})

	def _getFact(self, match):
		(key, ) = match.groups()
		(key, channel) = self.stripChannel(key)
		assert key, 'getFact: key can\' be empty'
		self.parser.update({
			'key':key,
			'channel':channel,
			})
		return True

	def _addFact(self, match):
		(key, s, value) = match.groups()
		(key, channel) = self.stripChannel(key)
		if s == 'es':
			value = '%s %s %s' % (key, s, value)
		assert key and value, 'addFact: key and value can\'t be empty'
		self.parser.update({
			'key':key,
			'channel':channel,
			'fact':value,
			})
		return True

	def _editFact(self, match):
		(edit, ) = match.groups()
		assert edit, 'editFact: edit can\'t be empty'
		parser = self.parser.copy()
		self.parse(edit, testMethods=('addFact', 'addAlias'))
		self.parser.update(parser)
		return True

	def _addAlias(self, match):
		(alias, fact) = match.groups()
		(alias, channel) = self.stripChannel(alias)
		assert alias and fact, 'addAlist: alias and fact can\'t be empty'
		self.parser.update({
			'key':alias,
			'channel':channel,
			'fact':'<alias>' + fact,
			})
		return True

	def _pipe(self, match):
		(msg, s, toNick) = match.groups()
		assert toNick and s and msg, 'pipe: there\'s an empty string'
		self.msg = msg
		if s == '>':
			private = True
		else:
			private = False
		self.parser.update({
				'toNick':toNick,
				'private':private,
				})

# function for save some hassle when creating a User instance from an id in a fact
def defUserById(id, fact):
	return User(user_id=id, inChannel=fact.inChannel, db=fact.db)

# class for group all bot's replies
class Replies:

	def __init__(self, prefix):
		self.prefix = prefix
		self.channelString = '%s en el canal %s'

	def p(self, s):
		return self.prefix + s

	def fact(self, irc, fact):
		key = self.p(fact.key)
		prefix = False
		if fact.toNick and fact.inChannel:
			if fact.toNick == str(conf.supybot.nick):
				irc.reply('Yo ya se que es %s, gracias.' % key, prefixNick=True)
				return
			usuarios = irc.state.channels[fact.inChannel].users
			if fact.toNick not in usuarios:
				irc.reply('No veo a nadie llamado %s' % fact.toNick, prefixNick=True)
				return
			if not fact.private:
				prefix = True
		else:
			fact.private = False
		if 'info' in fact.options:
			count = fact.request_count
			author = defUserById(fact.created_by, fact)
			author.pull()
			dateCreated = strTime(fact.created_at)
			if fact.edited_by != None:
				editor = defUserById(fact.edited_by, fact)
				editor.pull()
				dateEdited = strTime(fact.edited_at)
			else:
				editor = None
			if count > 1:
				sCount = 'se usó %s veces' % count
			elif count == 1:
				sCount = 'se usó una vez'
			else:
				sCount = 'no se usó'
			if fact.isAlias():
				reply = '\'%s\' es un alias a \'%s\' y fué creado por %s el ' \
				        '%s, %s.' % (fact.key, fact.fact[7:], author.getNick(),
				        dateCreated, sCount)
			else:
				reply = 'El facto \'%s\' fué creado por %s el %s, %s y tiene %s ' \
				        'carácteres.' % \
				        (fact.key, author.getNick(), dateCreated, sCount, len(fact))
			if fact.edited_by != None:
				reply += ' Se editó por última vez el %s por %s.' % \
				(dateEdited, editor.getNick())
			if fact.isDisabled():
				if fact.isLocked():
					reply += ' \'%s\' está bloqueado.' % fact.key
				else:
					reply += ' \'%s\' está borrado.' % fact.key
			else:
				if fact.isLocked():
					reply += ' \'%s\' está protegido contra cambios.' % fact.key
			if fact.channel:
				reply += ' Canal %s.' % fact.channel
			irc.reply(reply, to=fact.toNick, prefixNick=prefix, private=fact.private)
		elif 'raw' in fact.options:
			irc.reply(fact.raw(), to=fact.toNick, prefixNick=prefix, private=fact.private)
		else:
			irc.reply(fact, to=fact.toNick, prefixNick=prefix, private=fact.private)

	def addFactSuccess(self, irc, fact):
		key = self.p(fact.key)
		if fact.channel:
			key = self.channelString % (key, fact.channel)
		irc.reply('Facto %s creado.' % key)

	def factDuplicate(self, irc, fact):
		key = self.p(fact.key)
		if fact.channel:
			key = self.channelString % (key, fact.channel)
		irc.reply('El facto %s ya existe. Si desea editarlo utilise %sno.' % \
			(key, self.prefix))

	def delFactSuccess(self, irc, fact):
		key = self.p(fact.key)
		if fact.channel:
			key = self.channelString % (key, fact.channel)
		irc.reply('Facto %s borrado.' % key)

	def undelFactSuccess(self, irc, fact):
		irc.reply('Facto %s restaurado.' % self.p(fact.key))

	def editFactSuccess(self, irc, fact):
		key = self.p(fact.key)
		if fact.channel:
			key = self.channelString % (key, fact.channel)
		irc.reply('Facto %s actualizado.' % key)

	def unlockFactSuccess(self, irc, fact):
		irc.reply('Ahora el facto %s se puede editar.' % self.p(fact.key))

	def lockFactSuccess(self, irc, fact):
		irc.reply('Facto %s protegido contra cambios.' % self.p(fact.key))

	def factNotFound(self, irc, fact):
		irc.reply('El facto %s no existe.' % self.p(fact.key))

	def factInvalid(self, irc, fact):
		if len(fact):
			if len(fact.key) > fact.keyMaxLen:
				irc.reply('%s tiene %d carácteres. El máximo es %d' % (fact.key,
						len(fact.key), fact.keyMaxLen))
			elif len(fact) > fact.factMaxLen:
				irc.reply('Me diste un facto de %d carácteres. El máximo es %d' % \
					(len(fact), fact.factMaxLen))
			else:
				raise callbacks.Error, 'El facto es inválido: %s %s' % \
					(fact.key, fact.created_by)
		else:
			irc.reply('El facto %s está vacío.' % self.p(fact.key))

	def factForbidden(self, irc, fact):
		irc.reply('El facto %s está bloqueado.' % self.p(fact.key))

	def factLocked(self, irc, fact):
		irc.reply('El facto %s está protegido.' % self.p(fact.key))

	def searchResult(self, irc, keywords, facts):
		if facts:
			if len(facts) > 1:
				reply = 'Se encontraron %s factos: ' % len(facts)
			else:
				reply = 'Se encontró un facto: '
			reply += ' '.join([ '%s%s' % (self.prefix, v) for v in facts])
			irc.reply(reply)
		else:
			irc.reply('No se encontró nada con la clave \'%s\'' % str(keywords))


factosDb = plugins.DB('FactosDB', {'sqlite':SqliteDB})
class Factos(callbacks.Plugin):
	"""See README.txt for detailed info of how to use this plugin."""

	def __init__(self, irc):
		self.__parent = super(Factos, self)
		self.__parent.__init__(irc)
		self.db = factosDb()
		self.parser = ParseMsg()
		self.reply = Replies('!') # XXX get from config

	def getChannel(self, msg):
		channel = plugins.getChannel(msg.args[0])
		if channel.startswith('#'):
			return channel
		else:
			return None

	def createFact(self, msg, key='', **kwargs):
		channel = self.getChannel(msg)
		fact = Fact(key, channel, user=User(msg.prefix, channel))
		fact.db = self.db
		fact.user.db = self.db
		fact.update(kwargs)
		return fact

	def invalidCommand(self, irc, msg, tokens):
		message = ' '.join(tokens)
		fact = self.createFact(msg)
		parser = self.parser.parse(message)
		if not parser:
			return
		fact.update(parser)
		if hasattr(self, '_' + fact.method):
			log.info('%s called by %s at %s.' % (fact.method, fact.user.host, fact.inChannel))
			getattr(self, '_' + fact.method)(irc, fact)

	def _getFact(self, irc, fact):
		if fact.options:
			# we might want info about a deleted fact
			fact.pull(True)
		else:
			fact.pull()
		if fact:
			if fact.options:
				# don't check for alias or update fact if asking for info
				pass
			elif fact.isAlias():
				alias = fact.copy()
				fact.setToAlias()
				fact.pull()
				if not fact or fact.isDisabled():
					self.reply.factNotFound(irc, fact)
					return
				alias.setLastRequest()
				alias.push()
			else:
				fact.setLastRequest()
				fact.push()
			self.reply.fact(irc, fact)
			return
		self.reply.factNotFound(irc, fact)

	def _addFact(self, irc, fact):
		if fact.isAlias():
			self._addAlias(irc, fact)
			return
		copy = fact.copy()
		fact.pull(True)
		if fact:
			if fact.isDisabled():
				if fact.isLocked():
					self.reply.factForbidden(irc, fact)
					return
				# fact exists, but is disabled and editable
				fact.edit(copy.fact)
				if fact.isValid:
					fact.undisable()
					fact.push()
					self.reply.addFactSuccess(irc, fact)
				else:
					self.reply.factInvalid(irc, fact)
			else:
				self.reply.factDuplicate(irc, fact)
		else:
			if copy.isValid():
				copy.add()
				self.reply.addFactSuccess(irc, copy)
			else:
				self.reply.factInvalid(irc, copy)

	def _editFact(self, irc, fact):
		copy = fact.copy()
		fact.pull()
		if fact:
			if fact.isLocked():
				self.reply.factLocked(irc, fact)
				return
			fact.edit(copy.fact)
			if fact.isValid():
				fact.push()
				self.reply.editFactSuccess(irc, fact)
			else:
				self.reply.factInvalid(irc, fact)
		else:
			self.reply.factNotFound(irc, fact)

	def _addAlias(self, irc, fact):
		target = fact.copy()
		target.setToAlias()
		target.pull()
		if target.isAlias():
			target.setToAlias()
			target.pull()
		if not target:
			self.reply.factNotFound(irc, target)
			return
		fact.pull(True)
		if fact:
			if fact.isDisabled():
				if fact.isLocked():
					self.reply.factForbidden(irc, fact)
					return
				fact.edit('<alias>' + target.key)
				fact.undisable()
				if fact.isValid():
					fact.push()
					self.reply.addFactSuccess(irc, fact)
				else:
					self.reply.factInvalid(irc, fact)
			else:
				self.reply.factDuplicate(irc, fact)
		else:
			fact.fact = '<alias>' + target.key
			if fact.isValid():
				fact.add()
				self.reply.addFactSuccess(irc, fact)
			else:
				self.reply.factInvalid(irc, fact)

	def delFact(self, irc, msg, args, text):
		"""<facto>
		Borra el facto suministrado, note que no de elimina de la base de datos,
		puede restaurarlo con el comando desborrar."""
		parser = self.parser.parse(text, testMethods=('getFact', ), methodArgs={})
		if not parser:
			return
		fact = self.createFact(msg)
		fact.update(parser)
		fact.pull()
		if fact:
			if fact.isLocked():
				self.reply.factLocked(irc, fact)
				return
			fact.disable()
			fact.push()
			self.reply.delFactSuccess(irc, fact)
		else:
			self.reply.factNotFound(irc, fact)

	def undelFact(self, irc, msg, args, text):
		"""<facto>
		Restaura el facto anteriormente borrado."""
		parser = self.parser.parse(text, testMethods=('getFact', ), methodArgs={})
		if not parser:
			return
		fact = self.createFact(msg)
		fact.update(parser)
		fact.pull(True)
		if fact:
			if fact.isLocked():
				if fact.isDisabled():
					self.reply.factForbidden(irc, fact)
				else:
					self.reply.factLocked(irc, fact)
				return
			fact.undisable()
			fact.push()
			self.reply.undelFactSuccess(irc, fact)
		else:
			self.reply.factNotFound(irc, fact)

	def lockFact(self, irc, msg, args, text):
		"""<facto>
		Protege el facto contra modificaciones,
		necesitas estar registrado como admin del bot para usar este comando."""
		parser = self.parser.parse(text, testMethods=('getFact', ), methodArgs={})
		if not parser:
			return
		fact = self.createFact(msg)
		fact.update(parser)
		fact.pull(True)
		if fact:
			fact.lock()
			fact.push()
			self.reply.lockFactSuccess(irc, fact)
		else:
			self.reply.factNotFound(irc, fact)

	def unlockFact(self, irc, msg, args, text):
		"""<facto>
		Habilita el facto para ser modificado,
		necesitas estar registrado como admin del bot para usar este comando."""
		parser = self.parser.parse(text, testMethods=('getFact', ), methodArgs={})
		if not parser:
			return
		fact = self.createFact(msg)
		fact.update(parser)
		fact.pull(True)
		if fact:
			fact.unlock()
			fact.push()
			self.reply.unlockFactSuccess(irc, fact)
		else:
			self.reply.factNotFound(irc, fact)

	def search(self, irc, msg, args, text):
		"""<clave> [<clave> [<clave> ...]]
		Busca factos que contengan las claves suministradas."""
		keywords = text.split()
		inChannel = self.getChannel(msg)
		facts = self.db.search(keywords, inChannel)
		self.reply.searchResult(irc, keywords, facts)

	def editor(self, irc, msg, args, id):
		"""<número>
		Muestra información sobre el usuario identificado por <número> en la tabla de usuarios."""
		channel = self.getChannel(msg)
		user = User(user_id=int(id), inChannel=channel, db=self.db)
		user.pull()
		if user.host:
			irc.reply(user.raw())
	
	editor = wrap(editor, ['int'])


	buscar = wrap(search, ['text'])
	borrar = wrap(delFact, ['text'])
	desborrar = wrap(undelFact, ['text'])
	proteger = wrap(lockFact, [('checkCapability', 'admin'), 'something'])
	desproteger = wrap(unlockFact, [('checkCapability', 'admin'), 'something'])

Class = Factos

if __name__ == '__main__':
	pass

# I *hate* python code with spaces instead of tabs.
# vim:set shiftwidth=4 tabstop=4 noexpandtab textwidth=109:
