# -*- Encoding: UTF-8 -*-
###
# Copyright (c) 2008-2009, Elián Hanisch
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###

from supybot.test import *

Factos = plugin.loadPluginModule('Factos')
Fconf = conf.supybot.plugins.Factos
Fact = Factos.plugin.Fact


class DummyReplies:

	def __init__(self, repliesInstanse):
		self.replies = repliesInstanse

	def reply(self, msg, to=None, private=False, prefixNick=False):
		if prefixNick and to:
			msg = '%s: %s' % (to, msg)
		self.msg = msg

	def get(self, type, fact):
		getattr(self.replies, type)(self, fact)
		return self.msg


reply = DummyReplies(Factos.plugin.Replies('!'))
class FactosTestCase(ChannelPluginTestCase):
	plugins = ('Factos', )

	def setUp(self):
		super(FactosTestCase, self).setUp()
		self.irc.state.channels[self.channel].addUser('user1')
		reply.state = self.irc.state
		self.prefix = 'test!user@factos.net'

	def testAddFact(self):
		fact = Fact('foobar')
		self.assertResponse('foobar', reply.get('factNotFound', fact))
		self.assertResponse('foobar es una prueba', reply.get('addFactSuccess', fact))
		self.assertResponse('foobar', 'foobar es una prueba')
		self.assertResponse('foobar: ...', reply.get('factDuplicate', fact))
		fact1 = Fact('foo')
		self.assertResponse('foo:   bar', reply.get('addFactSuccess', fact1))
		self.assertResponse('foo', 'bar')
		#charlimits
		s = 'a' * 401
		longfact=Fact('longfact', fact=s)
		self.assertResponse('longfact:%s' % s , reply.get('factInvalid', longfact))
		s = 'a' * 26
		longfact=Fact(s, fact='asd')
		self.assertResponse('%s:asd' % s, reply.get('factInvalid', longfact))
		# invalid facts
		self.assertError(' test :  ')
		self.assertError('test es  ')
		self.assertError('  : test ')
		self.assertError('  es test ')
		# add a disabled fact
		self.assertNotError('borrar foo')
		self.assertResponse('foo es otro foo', reply.get('addFactSuccess', fact1))
		# add a fact with quotes
		conf.supybot.commands.quotes.set('')
		quote = Fact('comillas')
		self.assertResponse('comillas es "dobles" y  \'simples\'',
				reply.get('addFactSuccess', quote))
		self.assertResponse('comillas', 'comillas es "dobles" y \'simples\'')

	def testGetFact(self):
		fact = Fact('foo')
		self.assertResponse('foo', reply.get('factNotFound', fact))
		self.assertNotError('foo es bar')
		self.assertResponse('foo', 'foo es bar')

	def testEditFact(self):
		fact = Fact('foo')
		self.assertResponse('no, foo es bar', reply.get('factNotFound', fact))
		self.assertNotError('foo:bar')
		self.assertResponse('foo', 'bar')
		self.assertResponse('no, foo es bar', reply.get('editFactSuccess', fact))
		self.assertResponse('foo', 'foo es bar')
		self.assertResponse('no foo : bar', reply.get('editFactSuccess', fact))
		self.assertResponse('foo', 'bar')

	def testIgnoreCase(self):
		fact = Fact('test')
		self.assertNotError('tEsT:testing...')
		self.assertResponse('test', 'testing...')
		self.assertResponse('TEST', 'testing...')
		self.assertResponse('TEST:testing...', reply.get('factDuplicate', fact))
		self.assertNotError('PrUeBA es una prueba')
		self.assertResponse('prueba', 'PrUeBA es una prueba') # keeps case in reply
		self.assertNotError('foo es bar')
		self.assertResponse('FOO', 'foo es bar')
		self.assertNotError('BAR es foo')
		self.assertResponse('bar', 'BAR es foo')

	def testSpecialChars(self):
		self.assertNotError('ña_-a.z es válido')
		self.assertResponse('ña_-a.z', 'ña_-a.z es válido')

	def testRemoveFact(self):
		fact = Fact('foo')
		self.assertNotError('foo:bar')
		self.assertResponse('foo', 'bar')
		self.assertResponse('borrar foo', reply.get('delFactSuccess', fact))
		self.assertResponse('foo', reply.get('factNotFound', fact))
		self.assertResponse('borrar foo', reply.get('factNotFound', fact))
		self.assertResponse('desborrar foo', reply.get('undelFactSuccess', fact))
		self.assertResponse('foo', 'bar')
		self.assertResponse('desborrar foo', reply.get('undelFactSuccess', fact))
		self.assertResponse('foo', 'bar')

	def testLockFact(self):
		fact = Fact('foo')
		self.assertNotError('foo:bar')
		self.assertResponse('foo', 'bar')
		self.assertResponse('proteger foo', reply.get('lockFactSuccess', fact))
		self.assertResponse('no, foo es bar', reply.get('factLocked', fact))
		self.assertResponse('proteger foo', reply.get('lockFactSuccess', fact))
		self.assertResponse('borrar foo', reply.get('factLocked', fact))
		self.assertResponse('desproteger foo', reply.get('unlockFactSuccess', fact))
		self.assertResponse('no, foo es bar', reply.get('editFactSuccess', fact))
		self.assertResponse('desproteger foo', reply.get('unlockFactSuccess', fact))
		self.assertResponse('borrar foo', reply.get('delFactSuccess', fact))
		self.assertResponse('proteger foo', reply.get('lockFactSuccess', fact))
		self.assertResponse('desborrar foo', reply.get('factForbidden', fact))
		self.assertResponse('foo:bar', reply.get('factForbidden', fact))

	def testAlias(self):
		fact = Fact('foo')
		alias = Fact('test')
		alias1 = Fact('prueba')
		self.assertNotError('foo:bar')
		self.assertResponse('test alias foo', reply.get('addFactSuccess', alias))
		self.assertResponse('test', 'bar')
		self.assertResponse('prueba alias test', reply.get('addFactSuccess', alias1))
		self.assertResponse('prueba', 'bar')
		self.assertNotError('borrar foo')
		self.assertResponse('test', reply.get('factNotFound', fact))
		self.assertResponse('prueba', reply.get('factNotFound', fact))
		self.assertResponse('foo alias test', reply.get('factNotFound', fact))
		self.assertResponse('foo:<alias>test', reply.get('factNotFound', fact))
		self.assertNotError('desborrar foo')
		self.assertResponse('foo alias test', reply.get('factDuplicate', fact))
		self.assertNotError('no, foo alias test')
		self.assertResponse('foo', '<alias>foo')
		# create alias manually
		self.assertNotError('foobar es una prueba')
		self.assertNotError('test1:<alias>foobar')
		self.assertResponse('test1', 'foobar es una prueba')
		# create alias by editing
		self.assertNotError('no, foo alias foobar')
		self.assertResponse('foo', 'foobar es una prueba')
		self.assertRegexp('foo --raw', 'fact:<alias>foobar')

	def testChannelSpecificFacts(self):
		fact = Fact('foo', channel='#test')
		fact1 = Fact('bar', channel='#test')
		self.assertNotError('foo:global')
		self.assertResponse('foo', 'global')
		self.assertResponse('foo#test : local', reply.get('addFactSuccess', fact))
		self.assertResponse('foo', 'local')
		self.assertResponse('foo#test', 'local')
		self.assertResponse('foo#', 'global')
		self.assertResponse('borrar foo#test', reply.get('delFactSuccess', fact))
		self.assertResponse('foo', 'global')
		self.assertResponse('bar#test alias foo', reply.get('addFactSuccess', fact1))
		self.assertResponse('bar', 'global')
		self.assertResponse('bar#', reply.get('factNotFound', fact1))
		self.assertNotError('desborrar foo#test')
		self.assertResponse('bar', 'local')
		# channel edits
		m = self.getMsg('no, foo : local2')
		self.assertResponse('foo', 'local2')
		self.assertResponse('foo#test', 'local2')
		self.assertResponse('foo#', 'global')
		m = self.getMsg('no, foo#: global2')
		self.assertResponse('foo#', 'global2')
		m = self.getMsg('no, foo#test: local3')
		self.assertResponse('foo', 'local3')

	def testCounters(self):
		#fact = Fact('foo')
		m = self.getMsg('foo:bar')
		self.assertRegexp('foo --raw', 'request_count:0')
		for i in range(10):
			m = self.getMsg('foo')
		self.assertRegexp('foo --raw', 'request_count:10')
		self.assertRegexp('editor 0', 'facts:1')
		for i in range(10):
			m = self.getMsg('foo%s:bar' % i)
		self.assertRegexp('editor 0', 'facts:11 edits:0')
		for i in range(10):
			m = self.getMsg('no, foo:bar%s' % i)
		self.assertRegexp('editor 0', 'edits:10')
		self.prefix = 'moo!user@factos.net'
		m = self.getMsg('no, foo:bar')
		self.assertRegexp('editor 1', 'facts:0 edits:1')

	def testGetFactRaw(self):
		#fact = Fact('foo', fact='bar')
		m = self.getMsg('foo:bar')
		self.assertRegexp('foo --raw', r'key:foo created_by:0 ' \
				'created_at:\d+ edited_by:(\d+|None) edited_at:(\d+|None) ' \
				'flags:[0-3] flags_set_by:(\d+|None) flags_set_at:(\d+|None) '\
				'last_request_by:(.+?!.+?@.+?|None) ' \
				'last_request_at:(\d+|None) request_count:\d+ fact:bar')

	def testGetFactInfo(self):
		m = self.getMsg('foo:bar')
		m = self.getMsg('bar alias foo')
		self.assertRegexp('foo --info', r'El facto \'foo\' fué creado por ' \
			'test el \d\d/\d\d/\d\d, no se usó y tiene \d+ carácteres\.')
		self.assertRegexp('bar --info', r'\'bar\' es un alias a \'foo\' y ' \
			'fué creado por test el \d\d/\d\d/\d\d, no se usó\.')
		m = self.getMsg('no, foo:bar1')
		self.assertRegexp('foo --info', r'Se editó por última vez el ' \
			'\d\d/\d\d/\d\d por test\.')
		m = self.getMsg('proteger foo')
		self.assertRegexp('foo --info', r'\'foo\' está protegido contra cambios\.')
		m = self.getMsg('desproteger foo')
		m = self.getMsg('borrar foo')
		self.assertRegexp('foo --info', r'\'foo\' está borrado\.')
		m = self.getMsg('proteger foo')
		self.assertRegexp('foo --info', r'\'foo\' está bloqueado\.')


	def testEditorRaw(self):
		#fact = Fact('foo', fact='bar')
		m = self.getMsg('foo:bar')
		self.assertRegexp('editor 0', r'user_id:0 host:test!user@factos.net ' \
				'first_seen:\d+ last_seen:\d+ facts:\d+ edits:\d+')
		self.prefix = 'moo!user@factos.net'
		m = self.getMsg('foo1:bar')
		self.assertRegexp('editor 1', r'user_id:1 host:moo!user@factos.net ' \
				'first_seen:\d+ last_seen:\d+ facts:\d+ edits:\d+')

	def testPipe(self):
		fact = Fact('foo', fact='bar')
		fact.inChannel = self.channel
		self.getMsg('foo : bar')
		self.assertResponse('foo', 'bar')
		self.assertResponse('foo|user1', 'user1: bar')
		self.assertResponse('foo>user1', 'bar') # dunno how test if is sent
		                                        # privately
		fact.toNick = 'test'
		self.assertResponse('foo|test', reply.get('fact', fact))
		fact.toNick = 'user2'
		self.assertResponse('foo|user2', reply.get('fact', fact))



# I *hate* python code with spaces instead of tabs.
# vim:set shiftwidth=4 tabstop=4 noexpandtab textwidth=109:
